!******************************************************************************!
!                                                                              !
!    class_particles.f90                                                       !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016-2018 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!
!                                                           .---.              !
!                                                          /  .  \             !
!                                                         |\_/|   |            !
!                                                         |   |  /|            !
!          .----------------------------------------------------' |            !
!         /  .-.   class_particles.f90                            |            !
!        |  /   \  This code was generated with REPT1L.           |            !
!        | |\_.  | Change this file and you get what you deserve. |            !
!        |\|  | /| Author: J.N. Lang                              |            !
!        | `---' | UFO date: 03/2019                              |            !
!        |       | UFO version: 1.0                               |            !
!        |       | Generation date: 2019-10-14 13:42             /             !
!        |       |----------------------------------------------'              !
!        \       |                                                             !
!         \     /                                                              !
!          `---'                                                               !
!******************************************************************************!
module class_particles
  use constants_mdl
  use input_mdl, only: masscut,error_mdl,warning_mdl,nx,muMS
  implicit none
  save
  logical :: particle_model_init = .false.
  character(len=*), parameter :: modelname = "SM (QCD) + HEFT"
  character(len=*), parameter :: modelgauge = "'t Hooft-Feynman"
  character(len=*), parameter :: driver_timestamp = "Mon Oct 14 13:28:14 2019"
! number of particles in the model
  integer, parameter :: n_particles = 20
  integer, parameter :: n_orders = 2
  character(len=3), parameter, dimension(n_orders) :: order_ids                &
                  = (/'QCD', 'QED'/)

! Particle:\gamma, pdg_code:22
  integer, parameter :: P_A = 1
! Particle:Z, pdg_code:23
  integer, parameter :: P_Z = 2
! Particle:W^+, pdg_code:24
  integer, parameter :: P_WPlus = 3
! Particle:W^-, pdg_code:-24
  integer, parameter :: P_WMinus = 4
! Particle:g, pdg_code:21
  integer, parameter :: P_g = 5
! Particle:u_{G-}, pdg_code:82
  integer, parameter :: P_ghG = 6
! Particle:\bar{u}_{G}, pdg_code:-82
  integer, parameter :: P_ghGAnti = 7
! Particle:u, pdg_code:2
  integer, parameter :: P_u = 8
! Particle:\bar{u}, pdg_code:-2
  integer, parameter :: P_uAnti = 9
! Particle:c, pdg_code:4
  integer, parameter :: P_c = 10
! Particle:\bar{c}, pdg_code:-4
  integer, parameter :: P_cAnti = 11
! Particle:t, pdg_code:6
  integer, parameter :: P_t = 12
! Particle:\bar{t}, pdg_code:-6
  integer, parameter :: P_tAnti = 13
! Particle:d, pdg_code:1
  integer, parameter :: P_d = 14
! Particle:\bar{d}, pdg_code:-1
  integer, parameter :: P_dAnti = 15
! Particle:s, pdg_code:3
  integer, parameter :: P_s = 16
! Particle:\bar{s}, pdg_code:-3
  integer, parameter :: P_sAnti = 17
! Particle:b, pdg_code:5
  integer, parameter :: P_b = 18
! Particle:\bar{b}, pdg_code:-5
  integer, parameter :: P_bAnti = 19
! Particle:H, pdg_code:25
  integer, parameter :: P_H = 20

! Masses and widths
  real(kind=dp) :: MC = 0d0
  real(kind=dp) :: MB = 0d0
  real(kind=dp) :: MD = 0d0
  real(kind=dp) :: MH = 125d0
  real(kind=dp) :: MW = 80.357973609878d0
  real(kind=dp) :: MS = 0d0
  real(kind=dp) :: MU = 0d0
  real(kind=dp) :: MT = 173.2d0
  real(kind=dp) :: MZ = 91.153480619183d0
  real(kind=dp) :: WC = 0d0
  real(kind=dp) :: WB = 0d0
  real(kind=dp) :: WH = 0d0
  real(kind=dp) :: WW = 2.0842989982782d0
  real(kind=dp) :: WT = 0d0
  real(kind=dp) :: WZ = 2.4942663787728d0

  real(kind=dp) :: MC2, MB2, MD2, MH2, MW2, MS2, MU2, MT2, MZ2
! Unique mass ids
  integer :: nMC, nMB, nMD, nMH, nMW, nMS, nMU, nMT, nMZ
! Mass regularization
  integer :: MC_reg, MB_reg, MD_reg, MH_reg, MW_reg, MS_reg, MU_reg, MT_reg,   &
             MZ_reg
! The number of distinct mass regulators.
  integer :: n_masses
! Complex mass parameters
  complex(kind=dp) :: cMC, cMB, cMD, cMH, cMW, cMS, cMU, cMT, cMZ
  complex(kind=dp) :: cMC_reg, cMB_reg, cMD_reg, cMH_reg, cMW_reg, cMS_reg,    &
                cMU_reg, cMT_reg, cMZ_reg
  complex(kind=dp) :: cMC2, cMB2, cMD2, cMH2, cMW2, cMS2, cMU2, cMT2, cMZ2
  complex(kind=dp) :: cMC2_reg, cMB2_reg, cMD2_reg, cMH2_reg, cMW2_reg,        &
                cMS2_reg, cMU2_reg, cMT2_reg, cMZ2_reg
  complex(kind=dp) :: rMC, rMB, rMD, rMH, rMW, rMS, rMU, rMT, rMZ
  complex(kind=dp) :: rMC2, rMB2, rMD2, rMH2, rMW2, rMS2, rMU2, rMT2, rMZ2

! Resonances
  logical :: c_res = .false.
  logical :: b_res = .false.
  logical :: H_res = .false.
  logical :: W_res = .false.
  logical :: t_res = .false.
  logical :: Z_res = .false.

! Light particles
  logical :: MB_reg_light = .false.
  logical :: MU_reg_light = .false.
  logical :: MC_reg_light = .false.
  logical :: MD_reg_light = .false.
  logical :: MS_reg_light = .false.

! Maximal helicity configurations. Determined by the particle with highest spin  
! e.g. massive VB  -> 3
  integer, parameter :: max_polarizations = 3

! External parameter
  complex(kind=dp) :: aEW = 0.75552541674291d-02
  complex(kind=dp) :: aS = 0.118d0
  complex(kind=dp) :: yme = 0.000511d0
  complex(kind=dp) :: ymm = 0.10566d0
  complex(kind=dp) :: ymtau = 1.777d0
  complex(kind=dp) :: cgghfin = 0d0

! Internal parameter
  complex(kind=dp) :: ymc
  complex(kind=dp) :: ymb
  complex(kind=dp) :: ymt
  complex(kind=dp) :: ymdo
  complex(kind=dp) :: ymup
  complex(kind=dp) :: yms
  complex(kind=dp) :: ZERO
  complex(kind=dp) :: sw2
  complex(kind=dp) :: ee
  complex(kind=dp) :: cggh
  complex(kind=dp) :: gs
  complex(kind=dp) :: cw
  complex(kind=dp) :: sw
  complex(kind=dp) :: g1
  complex(kind=dp) :: gw
  complex(kind=dp) :: vev
  complex(kind=dp) :: lam
  complex(kind=dp) :: yb
  complex(kind=dp) :: yc
  complex(kind=dp) :: ydo
  complex(kind=dp) :: ye
  complex(kind=dp) :: ym
  complex(kind=dp) :: ys
  complex(kind=dp) :: yt
  complex(kind=dp) :: ytau
  complex(kind=dp) :: yup
  complex(kind=dp) :: muH2

! cmass2/cmass2_reg relates the mass id to the actual complex squared mass(_reg) 
! value.
  complex(kind=dp), allocatable :: cmass2(:), cmass2_reg(:)

  contains
    function sec(a) result(b)
      complex(kind=dp), intent(in) :: a
      complex(kind=dp) :: b
      b = 1/cos(a)
    end function sec

    function csc(a) result(b)
      complex(kind=dp), intent(in) :: a
      complex(kind=dp) :: b
      b = 1/sin(a)
    end function csc

    function cot(a) result(b)
      complex(kind=dp), intent(in) :: a
      complex(kind=dp) :: b
      b = 1/tan(a)
    end function cot

    function conjugate(a) result(b)
      complex(kind=dp), intent(in) :: a
      complex(kind=dp) :: b
      b = dconjg(a)
    end function conjugate

    function get_modelname_mdl() result(ret_modelname)
      character(len=100) :: ret_modelname
      ret_modelname = modelname
    end function get_modelname_mdl

    function get_modelgauge_mdl() result(ret_modelgauge)
      character(len=100) :: ret_modelgauge
      ret_modelgauge = modelgauge
    end function get_modelgauge_mdl

    function get_driver_timestamp_mdl() result(ret_driver_timestamp)
      character(len=200) :: ret_driver_timestamp
      ret_driver_timestamp = driver_timestamp
    end function get_driver_timestamp_mdl

    function get_cmass2_mdl(m_id) result(ret_cmass2)
      integer, intent(in) :: m_id
      complex(kind=dp) :: ret_cmass2
      ret_cmass2 = cmass2(m_id)
    end function get_cmass2_mdl

    function get_particle_cmass2_mdl(p_id) result(ret_cmass2)
      integer, intent(in) :: p_id
      integer :: m_id
      complex(kind=dp) :: ret_cmass2
      m_id = get_particle_mass_id_mdl(p_id)
      ret_cmass2 = cmass2(m_id)
    end function get_particle_cmass2_mdl

    function get_cmass2_reg_mdl(m_id) result(ret_cmass2_reg)
      integer, intent(in) :: m_id
      complex(kind=dp) :: ret_cmass2_reg
      ret_cmass2_reg = cmass2_reg(m_id)
    end function get_cmass2_reg_mdl

    function get_particle_cmass2_reg_mdl(p_id) result(ret_cmass2_reg)
      integer, intent(in) :: p_id
      integer :: m_id
      complex(kind=dp) :: ret_cmass2_reg
      m_id = get_particle_mass_id_mdl(p_id)
      ret_cmass2_reg = cmass2_reg(m_id)
    end function get_particle_cmass2_reg_mdl

    function get_n_masses_mdl() result(ret_n_masses)
      integer :: ret_n_masses
      ret_n_masses = n_masses
    end function get_n_masses_mdl

    function get_n_particles_mdl() result(ret_n_particles)
      integer :: ret_n_particles
      ret_n_particles = n_particles
    end function get_n_particles_mdl

    function get_order_id_mdl(o_id) result(ret_order)
      integer, intent(in) :: o_id
      character(len=3) :: ret_order
      ret_order = order_ids(o_id)
    end function get_order_id_mdl

    function get_n_orders_mdl() result(ret_n_orders)
      integer :: ret_n_orders
      ret_n_orders = size(order_ids)
    end function get_n_orders_mdl

    function get_max_polarizations_mdl() result(ret_max_polarizations)
      integer :: ret_max_polarizations
      ret_max_polarizations = max_polarizations
    end function get_max_polarizations_mdl

    function get_particle_polarizations_mdl(p) result(npolarizations)
      integer, intent(in) :: p
      integer :: npolarizations
      character(len=3) :: p_str
      select case(get_particle_mass_reg_mdl(p))
      case(1,2)
        select case(get_particle_spin_mdl(p))
        case(-1,1)
          npolarizations = 1
        case(2,3)
          npolarizations = 2
        case default
          write(p_str,'(i3)') p
          call error_mdl("Spin other than -1,1,2,3 not supported. P_id:" // &
                        trim(p_str), where='get_particle_polarizations_mdl')
        end select
      case(3)
        select case(get_particle_spin_mdl(p))
        case(-1,1)
          npolarizations = 1
        case(2)
          npolarizations = 2
        case(3)
          npolarizations = 3
        case default
          write(p_str,'(i3)') p
          call error_mdl("Spin other than -1,1,2,3 not supported. P_id:" // &
                        trim(p_str), where='get_particle_polarizations_mdl')
          stop
        end select
      case default
        write(p_str,'(i3)') p
        call error_mdl("Mass reg cannot be other than 1,2 or 3. P_id:" // &
                      trim(p_str), where='get_particle_polarizations_mdl')
        stop
      end select
    end function get_particle_polarizations_mdl

    function is_particle_model_init_mdl() result(ret_init)
      logical :: ret_init
      ret_init = particle_model_init
    end function is_particle_model_init_mdl

    subroutine set_particle_model_init_mdl(set_init)
      logical, intent(in) :: set_init
      particle_model_init = set_init
    end subroutine set_particle_model_init_mdl

    subroutine clear_particles_mdl
      if (allocated(cmass2_reg)) deallocate(cmass2_reg)
      if (allocated(cmass2_reg)) deallocate(cmass2_reg)
      particle_model_init = .false.
    end subroutine clear_particles_mdl

    subroutine set_masscut_mdl (m)
      ! Sets mass limit below which particles are considered massless.
      real(dp), intent(in) :: m

      masscut = m
      if (is_particle_model_init_mdl()) then
        call warning_mdl('Masscut needs to be set before defining processes.',&
                        where='set_masscut_mdl')
      end if

    end subroutine set_masscut_mdl

    subroutine set_light_particle_mdl(p)
      character(*), intent(in) :: p
      call set_light_particle_value_mdl(p, .true.)
    end subroutine set_light_particle_mdl

    subroutine unset_light_particle_mdl(p)
      character(*), intent(in) :: p
      call set_light_particle_value_mdl(p, .false.)
    end subroutine unset_light_particle_mdl

    subroutine set_resonant_particle_mdl(p)
      character(*), intent(in) :: p
      call set_resonant_particle_value_mdl(p, .true.)
    end subroutine set_resonant_particle_mdl

    subroutine unset_resonant_particle_mdl(p)
      character(*), intent(in) :: p
      call set_resonant_particle_value_mdl(p, .false.)
    end subroutine unset_resonant_particle_mdl

    subroutine init_cmasses_mdl()
!     Init mass squared parameters
      MC2 = MC**2
      MB2 = MB**2
      MD2 = MD**2
      MH2 = MH**2
      MW2 = MW**2
      MS2 = MS**2
      MU2 = MU**2
      MT2 = MT**2
      MZ2 = MZ**2

!     Start with mass Ids and complex masses
!     The mass ids are given by their order in UFO
      n_masses = 2
!     Mass representative: Z
      MZ_reg = 3
      n_masses = n_masses + 1
      nMZ = n_masses
      if (Z_res .eqv. .false.) then
        cMZ2 = MZ**2 - cima*WZ*MZ
        cMZ2_reg = MZ**2 - cima*WZ*MZ
      else
        cMZ2 = cmplx(MZ**2, 0d0,kind=dp)
        cMZ2_reg = cmplx(MZ**2, 0d0,kind=dp)
      end if 

!     Mass representative: W+
      MW_reg = 3
      n_masses = n_masses + 1
      nMW = n_masses
      if (W_res .eqv. .false.) then
        cMW2 = MW**2 - cima*WW*MW
        cMW2_reg = MW**2 - cima*WW*MW
      else
        cMW2 = cmplx(MW**2, 0d0,kind=dp)
        cMW2_reg = cmplx(MW**2, 0d0,kind=dp)
      end if 

!     Mass representative: u
      if (MU .lt. masscut .or. MU_reg_light) then
        if (MU .eq. 0d0) then
          nMU = 1
          MU_reg = 1
          cMU2 = cnul
          cMU2_reg = cnul
        else
          MU_reg = 2
          n_masses = n_masses + 1
          nMU = n_masses
          cMU2 = cnul
          cMU2_reg = cmplx(MU**2, 0d0,kind=dp)
        end if 
      else
        MU_reg = 3
        n_masses = n_masses + 1
        nMU = n_masses
        cMU2 = cmplx(MU**2, 0d0,kind=dp)
        cMU2_reg = cmplx(MU**2, 0d0,kind=dp)
      end if 

!     Mass representative: c
      if (MC .lt. masscut .or. MC_reg_light) then
        if (MC .eq. 0d0) then
          nMC = 1
          MC_reg = 1
          cMC2 = cnul
          cMC2_reg = cnul
        else
          MC_reg = 2
          n_masses = n_masses + 1
          nMC = n_masses
          cMC2 = cnul
          cMC2_reg = MC**2 - cima*WC*MC
          if (abs(WC) .gt. 0d0) then
            call error_mdl('Finite width WC>0 not tested in massreg.', where   &
            ='init_cmasses_mdl')
          end if 
        end if 
      else
        MC_reg = 3
        n_masses = n_masses + 1
        nMC = n_masses
        if (c_res .eqv. .false.) then
          cMC2 = MC**2 - cima*WC*MC
          cMC2_reg = MC**2 - cima*WC*MC
        else
          cMC2 = cmplx(MC**2, 0d0,kind=dp)
          cMC2_reg = cmplx(MC**2, 0d0,kind=dp)
        end if 
      end if 

!     Mass representative: t
      if (MT .lt. masscut) then
        if (MT .eq. 0d0) then
          nMT = 1
          MT_reg = 1
          cMT2 = cnul
          cMT2_reg = cnul
        else
          MT_reg = 2
          n_masses = n_masses + 1
          nMT = n_masses
          cMT2 = cnul
          cMT2_reg = MT**2 - cima*WT*MT
          if (abs(WT) .gt. 0d0) then
            call error_mdl('Finite width WT>0 not tested in massreg.', where   &
            ='init_cmasses_mdl')
          end if 
        end if 
      else
        MT_reg = 3
        n_masses = n_masses + 1
        nMT = n_masses
        if (t_res .eqv. .false.) then
          cMT2 = MT**2 - cima*WT*MT
          cMT2_reg = MT**2 - cima*WT*MT
        else
          cMT2 = cmplx(MT**2, 0d0,kind=dp)
          cMT2_reg = cmplx(MT**2, 0d0,kind=dp)
        end if 
      end if 

!     Mass representative: d
      if (MD .lt. masscut .or. MD_reg_light) then
        if (MD .eq. 0d0) then
          nMD = 1
          MD_reg = 1
          cMD2 = cnul
          cMD2_reg = cnul
        else
          MD_reg = 2
          n_masses = n_masses + 1
          nMD = n_masses
          cMD2 = cnul
          cMD2_reg = cmplx(MD**2, 0d0,kind=dp)
        end if 
      else
        MD_reg = 3
        n_masses = n_masses + 1
        nMD = n_masses
        cMD2 = cmplx(MD**2, 0d0,kind=dp)
        cMD2_reg = cmplx(MD**2, 0d0,kind=dp)
      end if 

!     Mass representative: s
      if (MS .lt. masscut .or. MS_reg_light) then
        if (MS .eq. 0d0) then
          nMS = 1
          MS_reg = 1
          cMS2 = cnul
          cMS2_reg = cnul
        else
          MS_reg = 2
          n_masses = n_masses + 1
          nMS = n_masses
          cMS2 = cnul
          cMS2_reg = cmplx(MS**2, 0d0,kind=dp)
        end if 
      else
        MS_reg = 3
        n_masses = n_masses + 1
        nMS = n_masses
        cMS2 = cmplx(MS**2, 0d0,kind=dp)
        cMS2_reg = cmplx(MS**2, 0d0,kind=dp)
      end if 

!     Mass representative: b
      if (MB .lt. masscut .or. MB_reg_light) then
        if (MB .eq. 0d0) then
          nMB = 1
          MB_reg = 1
          cMB2 = cnul
          cMB2_reg = cnul
        else
          MB_reg = 2
          n_masses = n_masses + 1
          nMB = n_masses
          cMB2 = cnul
          cMB2_reg = MB**2 - cima*WB*MB
          if (abs(WB) .gt. 0d0) then
            call error_mdl('Finite width WB>0 not tested in massreg.', where   &
            ='init_cmasses_mdl')
          end if 
        end if 
      else
        MB_reg = 3
        n_masses = n_masses + 1
        nMB = n_masses
        if (b_res .eqv. .false.) then
          cMB2 = MB**2 - cima*WB*MB
          cMB2_reg = MB**2 - cima*WB*MB
        else
          cMB2 = cmplx(MB**2, 0d0,kind=dp)
          cMB2_reg = cmplx(MB**2, 0d0,kind=dp)
        end if 
      end if 

!     Mass representative: H
      if (MH .lt. masscut) then
        if (MH .eq. 0d0) then
          nMH = 1
          MH_reg = 1
          cMH2 = cnul
          cMH2_reg = cnul
        else
          MH_reg = 2
          n_masses = n_masses + 1
          nMH = n_masses
          cMH2 = cnul
          cMH2_reg = MH**2 - cima*WH*MH
          if (abs(WH) .gt. 0d0) then
            call error_mdl('Finite width WH>0 not tested in massreg.', where   &
            ='init_cmasses_mdl')
          end if 
        end if 
      else
        MH_reg = 3
        n_masses = n_masses + 1
        nMH = n_masses
        if (H_res .eqv. .false.) then
          cMH2 = MH**2 - cima*WH*MH
          cMH2_reg = MH**2 - cima*WH*MH
        else
          cMH2 = cmplx(MH**2, 0d0,kind=dp)
          cMH2_reg = cmplx(MH**2, 0d0,kind=dp)
        end if 
      end if 


      if (allocated(cmass2)) deallocate(cmass2)
      allocate(cmass2(n_masses))
      if (allocated(cmass2_reg)) deallocate(cmass2_reg)
      allocate(cmass2_reg(n_masses))
      cmass2(1:2) = cnul
      cmass2_reg(1:2) = cnul
      cmass2(nMC) = cMC2
      cmass2_reg(nMC) = cMC2_reg
      cmass2(nMB) = cMB2
      cmass2_reg(nMB) = cMB2_reg
      cmass2(nMD) = cMD2
      cmass2_reg(nMD) = cMD2_reg
      cmass2(nMH) = cMH2
      cmass2_reg(nMH) = cMH2_reg
      cmass2(nMW) = cMW2
      cmass2_reg(nMW) = cMW2_reg
      cmass2(nMS) = cMS2
      cmass2_reg(nMS) = cMS2_reg
      cmass2(nMU) = cMU2
      cmass2_reg(nMU) = cMU2_reg
      cmass2(nMT) = cMT2
      cmass2_reg(nMT) = cMT2_reg
      cmass2(nMZ) = cMZ2
      cmass2_reg(nMZ) = cMZ2_reg

      cMC = sqrt(cMC2)
      cMC_reg = sqrt(cMC2_reg)
      cMB = sqrt(cMB2)
      cMB_reg = sqrt(cMB2_reg)
      cMD = sqrt(cMD2)
      cMD_reg = sqrt(cMD2_reg)
      cMH = sqrt(cMH2)
      cMH_reg = sqrt(cMH2_reg)
      cMW = sqrt(cMW2)
      cMW_reg = sqrt(cMW2_reg)
      cMS = sqrt(cMS2)
      cMS_reg = sqrt(cMS2_reg)
      cMU = sqrt(cMU2)
      cMU_reg = sqrt(cMU2_reg)
      cMT = sqrt(cMT2)
      cMT_reg = sqrt(cMT2_reg)
      cMZ = sqrt(cMZ2)
      cMZ_reg = sqrt(cMZ2_reg)

      rMC2 = cmplx(real(cMC2),0d0,kind=dp)
      rMC = sqrt(rMC2)
      rMB2 = cmplx(real(cMB2),0d0,kind=dp)
      rMB = sqrt(rMB2)
      rMD2 = cmplx(real(cMD2),0d0,kind=dp)
      rMD = sqrt(rMD2)
      rMH2 = cmplx(real(cMH2),0d0,kind=dp)
      rMH = sqrt(rMH2)
      rMW2 = cmplx(real(cMW2),0d0,kind=dp)
      rMW = sqrt(rMW2)
      rMS2 = cmplx(real(cMS2),0d0,kind=dp)
      rMS = sqrt(rMS2)
      rMU2 = cmplx(real(cMU2),0d0,kind=dp)
      rMU = sqrt(rMU2)
      rMT2 = cmplx(real(cMT2),0d0,kind=dp)
      rMT = sqrt(rMT2)
      rMZ2 = cmplx(real(cMZ2),0d0,kind=dp)
      rMZ = sqrt(rMZ2)

      call init_parameters_mdl()

    end subroutine init_cmasses_mdl

    subroutine init_parameters_mdl()
      ymc = cMC
      ymb = cMB
      ymt = cMT
      ymdo = cMD
      ymup = cMU
      yms = cMS
      ZERO = 0d0
      sw2 = -cMW**2d0/cMZ**2d0 + 1d0
      ee = 2d0*sqrt(pi)*sqrt(aEW)
      cggh = aS*(cgghfin + 1d0)/(12d0*pi)
      gs = 2d0*sqrt(pi)*sqrt(aS)
      cw = sqrt(-sw2 + 1d0)
      sw = sqrt(sw2)
      g1 = -ee/(2d0*cw)
      gw = ee/sw
      vev = 2d0*cMW*sw/ee
      lam = 2d0*cMH**2d0/vev**2d0
      yb = sqrt(2d0)*ymb/vev
      yc = sqrt(2d0)*ymc/vev
      ydo = sqrt(2d0)*ymdo/vev
      ye = sqrt(2d0)*yme/vev
      ym = sqrt(2d0)*ymm/vev
      ys = sqrt(2d0)*yms/vev
      yt = sqrt(2d0)*ymt/vev
      ytau = sqrt(2d0)*ymtau/vev
      yup = sqrt(2d0)*ymup/vev
      muH2 = lam*vev**2d0/4d0
    end subroutine init_parameters_mdl

    subroutine init_particles_mdl()
      call init_cmasses_mdl
    end subroutine init_particles_mdl

    function get_doublet_partner_mdl(particle_in) result(particle_out)
      integer :: particle_in, particle_out
      select case(particle_in)
      case(P_tAnti)
        particle_out = P_bAnti
      case(P_cAnti)
        particle_out = P_sAnti
      case(P_d)
        particle_out = P_u
      case(P_b)
        particle_out = P_t
      case(P_c)
        particle_out = P_s
      case(P_uAnti)
        particle_out = P_dAnti
      case(P_bAnti)
        particle_out = P_tAnti
      case(P_dAnti)
        particle_out = P_uAnti
      case(P_sAnti)
        particle_out = P_cAnti
      case(P_t)
        particle_out = P_b
      case(P_u)
        particle_out = P_d
      case(P_s)
        particle_out = P_c
      case default
        particle_out = 0
      end select 
    end function get_doublet_partner_mdl

    function is_firstgen_mdl(particle) result(isFirstGen)
      integer :: particle
      logical :: isFirstGen
      select case(particle)
      case(P_d, P_dAnti, P_u, P_uAnti)
        isFirstGen = .true.
      case default
        isFirstGen = .false.
      end select
    end function is_firstgen_mdl

    function get_secondgen_mdl(particle) result(secondGen)
      integer :: particle, secondGen
      select case(particle)
      case(P_d)
        secondGen = P_s
      case(P_u)
        secondGen = P_c
      case(P_dAnti)
        secondGen = P_sAnti
      case(P_uAnti)
        secondGen = P_cAnti
      end select
    end function get_secondgen_mdl

    function get_thirdgen_mdl(particle) result(thirdGen)
      integer :: particle, thirdGen
      select case(particle)
      case(P_d)
        thirdGen = P_b
      case(P_u)
        thirdGen = P_t
      case(P_dAnti)
        thirdGen = P_bAnti
      case(P_uAnti)
        thirdGen = P_tAnti
      end select
    end function get_thirdgen_mdl

    function get_particle_id_mdl(p_str) result(particle_id)
      character(*), intent(in) :: p_str
      integer :: particle_id
      select case(p_str)
      case("A")
        particle_id = 1
      case("Z")
        particle_id = 2
      case("W+")
        particle_id = 3
      case("W-")
        particle_id = 4
      case("g")
        particle_id = 5
      case("ghG")
        particle_id = 6
      case("ghG~")
        particle_id = 7
      case("u")
        particle_id = 8
      case("u~")
        particle_id = 9
      case("c")
        particle_id = 10
      case("c~")
        particle_id = 11
      case("t")
        particle_id = 12
      case("t~")
        particle_id = 13
      case("d")
        particle_id = 14
      case("d~")
        particle_id = 15
      case("s")
        particle_id = 16
      case("s~")
        particle_id = 17
      case("b")
        particle_id = 18
      case("b~")
        particle_id = 19
      case("H")
        particle_id = 20
      end select 
    end function get_particle_id_mdl

    function get_particle_name_mdl(p) result(pname)
      integer, intent(in) :: p
      character(len=10) :: pname
      character(len=3) :: p_str
      select case(p)
      case(P_A)
        pname = "A"
      case(P_Z)
        pname = "Z"
      case(P_WPlus)
        pname = "W+"
      case(P_WMinus)
        pname = "W-"
      case(P_g)
        pname = "g"
      case(P_ghG)
        pname = "ghG"
      case(P_ghGAnti)
        pname = "ghG~"
      case(P_u)
        pname = "u"
      case(P_uAnti)
        pname = "u~"
      case(P_c)
        pname = "c"
      case(P_cAnti)
        pname = "c~"
      case(P_t)
        pname = "t"
      case(P_tAnti)
        pname = "t~"
      case(P_d)
        pname = "d"
      case(P_dAnti)
        pname = "d~"
      case(P_s)
        pname = "s"
      case(P_sAnti)
        pname = "s~"
      case(P_b)
        pname = "b"
      case(P_bAnti)
        pname = "b~"
      case(P_H)
        pname = "H"
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_name_mdl")
      end select 
    end function get_particle_name_mdl

    function get_particle_texname_mdl(p) result(ptexname)
      integer, intent(in) :: p
      character(len=30) :: ptexname
      character(len=3) :: p_str
      select case(p)
      case(P_A)
        ptexname = "\gamma"
      case(P_Z)
        ptexname = "Z"
      case(P_WPlus)
        ptexname = "W^+"
      case(P_WMinus)
        ptexname = "W^-"
      case(P_g)
        ptexname = "g"
      case(P_ghG)
        ptexname = "u_{G-}"
      case(P_ghGAnti)
        ptexname = "\bar{u}_{G}"
      case(P_u)
        ptexname = "u"
      case(P_uAnti)
        ptexname = "\bar{u}"
      case(P_c)
        ptexname = "c"
      case(P_cAnti)
        ptexname = "\bar{c}"
      case(P_t)
        ptexname = "t"
      case(P_tAnti)
        ptexname = "\bar{t}"
      case(P_d)
        ptexname = "d"
      case(P_dAnti)
        ptexname = "\bar{d}"
      case(P_s)
        ptexname = "s"
      case(P_sAnti)
        ptexname = "\bar{s}"
      case(P_b)
        ptexname = "b"
      case(P_bAnti)
        ptexname = "\bar{b}"
      case(P_H)
        ptexname = "H"
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_texname_mdl")
      end select 
    end function get_particle_texname_mdl

    function get_antiparticle_id_mdl(p_str) result(particle_id)
      character(*), intent(in) :: p_str
      integer :: particle_id
      select case(p_str)
      case("A")
        particle_id = 1
      case("Z")
        particle_id = 2
      case("W-")
        particle_id = 3
      case("W+")
        particle_id = 4
      case("g")
        particle_id = 5
      case("ghG~")
        particle_id = 6
      case("ghG")
        particle_id = 7
      case("u~")
        particle_id = 8
      case("u")
        particle_id = 9
      case("c~")
        particle_id = 10
      case("c")
        particle_id = 11
      case("t~")
        particle_id = 12
      case("t")
        particle_id = 13
      case("d~")
        particle_id = 14
      case("d")
        particle_id = 15
      case("s~")
        particle_id = 16
      case("s")
        particle_id = 17
      case("b~")
        particle_id = 18
      case("b")
        particle_id = 19
      case("H")
        particle_id = 20
      case default
        call error_mdl("Particle  "// trim(p_str) // " not defined.", where="  &
        get_antiparticle_id_mdl")
      end select 
    end function get_antiparticle_id_mdl

    function get_particle_antiparticle_mdl(p) result(ppa)
      integer, intent(in) :: p
      integer :: ppa
      character(len=3) :: p_str
      select case(p)
      case(P_A)
        ppa = P_A
      case(P_Z)
        ppa = P_Z
      case(P_WPlus)
        ppa = P_WMinus
      case(P_WMinus)
        ppa = P_WPlus
      case(P_g)
        ppa = P_g
      case(P_ghG)
        ppa = P_ghGAnti
      case(P_ghGAnti)
        ppa = P_ghG
      case(P_u)
        ppa = P_uAnti
      case(P_uAnti)
        ppa = P_u
      case(P_c)
        ppa = P_cAnti
      case(P_cAnti)
        ppa = P_c
      case(P_t)
        ppa = P_tAnti
      case(P_tAnti)
        ppa = P_t
      case(P_d)
        ppa = P_dAnti
      case(P_dAnti)
        ppa = P_d
      case(P_s)
        ppa = P_sAnti
      case(P_sAnti)
        ppa = P_s
      case(P_b)
        ppa = P_bAnti
      case(P_bAnti)
        ppa = P_b
      case(P_H)
        ppa = P_H
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_antiparticle_mdl")
      end select 
    end function get_particle_antiparticle_mdl

    function get_particle_spin_mdl(p) result(pspin)
      integer, intent(in) :: p
      integer :: pspin
      character(len=3) :: p_str
      select case(p)
      case(P_A)
        pspin = 3
      case(P_Z)
        pspin = 3
      case(P_WPlus)
        pspin = 3
      case(P_WMinus)
        pspin = 3
      case(P_g)
        pspin = 3
      case(P_ghG)
        pspin = -1
      case(P_ghGAnti)
        pspin = -1
      case(P_u)
        pspin = 2
      case(P_uAnti)
        pspin = 2
      case(P_c)
        pspin = 2
      case(P_cAnti)
        pspin = 2
      case(P_t)
        pspin = 2
      case(P_tAnti)
        pspin = 2
      case(P_d)
        pspin = 2
      case(P_dAnti)
        pspin = 2
      case(P_s)
        pspin = 2
      case(P_sAnti)
        pspin = 2
      case(P_b)
        pspin = 2
      case(P_bAnti)
        pspin = 2
      case(P_H)
        pspin = 1
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_spin_mdl")
      end select 
    end function get_particle_spin_mdl

    function get_particle_colour_mdl(p) result(pcolour)
      integer, intent(in) :: p
      integer :: pcolour
      character(len=3) :: p_str
      select case(p)
      case(P_A)
        pcolour = 1
      case(P_Z)
        pcolour = 1
      case(P_WPlus)
        pcolour = 1
      case(P_WMinus)
        pcolour = 1
      case(P_g)
        pcolour = 8
      case(P_ghG)
        pcolour = 8
      case(P_ghGAnti)
        pcolour = 8
      case(P_u)
        pcolour = 3
      case(P_uAnti)
        pcolour = -3
      case(P_c)
        pcolour = 3
      case(P_cAnti)
        pcolour = -3
      case(P_t)
        pcolour = 3
      case(P_tAnti)
        pcolour = -3
      case(P_d)
        pcolour = 3
      case(P_dAnti)
        pcolour = -3
      case(P_s)
        pcolour = 3
      case(P_sAnti)
        pcolour = -3
      case(P_b)
        pcolour = 3
      case(P_bAnti)
        pcolour = -3
      case(P_H)
        pcolour = 1
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_colour_mdl")
      end select 
    end function get_particle_colour_mdl

    function get_mass_name_mdl(mass_id) result(mass_name)
      integer, intent(in) :: mass_id
      character(len=6) :: mass_name, mass_id_str
      if (mass_id .le. 2) then
        mass_name = "M0"
      else if(mass_id .eq. get_particle_mass_id_mdl(2)) then
        mass_name = "MZ"
      else if(mass_id .eq. get_particle_mass_id_mdl(3)) then
        mass_name = "MW"
      else if(mass_id .eq. get_particle_mass_id_mdl(8)) then
        mass_name = "MU"
      else if(mass_id .eq. get_particle_mass_id_mdl(10)) then
        mass_name = "MC"
      else if(mass_id .eq. get_particle_mass_id_mdl(12)) then
        mass_name = "MT"
      else if(mass_id .eq. get_particle_mass_id_mdl(14)) then
        mass_name = "MD"
      else if(mass_id .eq. get_particle_mass_id_mdl(16)) then
        mass_name = "MS"
      else if(mass_id .eq. get_particle_mass_id_mdl(18)) then
        mass_name = "MB"
      else if(mass_id .eq. get_particle_mass_id_mdl(20)) then
        mass_name = "MH"
      else
        write(mass_id_str,'(i3)') mass_id
        call error_mdl("Mass id " // trim(mass_id_str) // " not defined.",     &
        where="get_mass_name_mdl")
      end if 
    end function get_mass_name_mdl

    function get_particle_mass_id_mdl(p) result(pmassid)
      integer, intent(in) :: p
      integer :: pmassid
      select case(p)
      case(P_A)
        pmassid = 1
      case(P_Z)
        pmassid = nMZ
      case(P_WPlus)
        pmassid = nMW
      case(P_WMinus)
        pmassid = nMW
      case(P_g)
        pmassid = 1
      case(P_ghG)
        pmassid = 1
      case(P_ghGAnti)
        pmassid = 1
      case(P_u)
        pmassid = nMU
      case(P_uAnti)
        pmassid = nMU
      case(P_c)
        pmassid = nMC
      case(P_cAnti)
        pmassid = nMC
      case(P_t)
        pmassid = nMT
      case(P_tAnti)
        pmassid = nMT
      case(P_d)
        pmassid = nMD
      case(P_dAnti)
        pmassid = nMD
      case(P_s)
        pmassid = nMS
      case(P_sAnti)
        pmassid = nMS
      case(P_b)
        pmassid = nMB
      case(P_bAnti)
        pmassid = nMB
      case(P_H)
        pmassid = nMH
      end select 
    end function get_particle_mass_id_mdl

    function get_particle_mass_reg_mdl(p) result(pmassreg)
      integer, intent(in) :: p
      integer :: pmassreg
      select case(p)
      case(P_A)
        pmassreg = 1
      case(P_Z)
        pmassreg = MZ_reg
      case(P_WPlus)
        pmassreg = MW_reg
      case(P_WMinus)
        pmassreg = MW_reg
      case(P_g)
        pmassreg = 1
      case(P_ghG)
        pmassreg = 1
      case(P_ghGAnti)
        pmassreg = 1
      case(P_u)
        pmassreg = MU_reg
      case(P_uAnti)
        pmassreg = MU_reg
      case(P_c)
        pmassreg = MC_reg
      case(P_cAnti)
        pmassreg = MC_reg
      case(P_t)
        pmassreg = MT_reg
      case(P_tAnti)
        pmassreg = MT_reg
      case(P_d)
        pmassreg = MD_reg
      case(P_dAnti)
        pmassreg = MD_reg
      case(P_s)
        pmassreg = MS_reg
      case(P_sAnti)
        pmassreg = MS_reg
      case(P_b)
        pmassreg = MB_reg
      case(P_bAnti)
        pmassreg = MB_reg
      case(P_H)
        pmassreg = MH_reg
      end select 
    end function get_particle_mass_reg_mdl

    function get_particle_width_mdl(p) result(pwidth)
      integer, intent(in) :: p
      complex(kind=dp) :: pwidth
      select case(p)
      case(P_Z)
        pwidth = WZ
      case(P_WPlus)
        pwidth = WW
      case(P_WMinus)
        pwidth = WW
      case(P_c)
        pwidth = WC
      case(P_cAnti)
        pwidth = WC
      case(P_t)
        pwidth = WT
      case(P_tAnti)
        pwidth = WT
      case(P_b)
        pwidth = WB
      case(P_bAnti)
        pwidth = WB
      case(P_H)
        pwidth = WH
      case default
        pwidth = cnul
      end select 
    end function get_particle_width_mdl

    subroutine set_parameter_mdl(param,value)
      character(*), intent(in) :: param
      complex(dp), intent(in) :: value
      if (particle_model_init) then
        call warning_mdl                                                       &
        ('Parameters need to be set before the process  generation.', where    &
        ='set_parameter_mdl')
      end if 
      select case(param)
        case("MB")
          MB = real(value,kind=dp)
        case("MT")
          MT = real(value,kind=dp)
        case("yme")
          yme = real(value,kind=dp)
        case("WW")
          WW = real(value,kind=dp)
        case("ymt")
          ymt = real(value,kind=dp)
        case("MH")
          MH = real(value,kind=dp)
        case("MW")
          MW = real(value,kind=dp)
        case("ymdo")
          ymdo = real(value,kind=dp)
        case("aEW")
          aEW = real(value,kind=dp)
        case("ymc")
          ymc = real(value,kind=dp)
        case("cgghfin")
          cgghfin = real(value,kind=dp)
        case("muMS")
          muMS = real(value,kind=dp)
        case("WZ")
          WZ = real(value,kind=dp)
        case("yms")
          yms = real(value,kind=dp)
        case("ymm")
          ymm = real(value,kind=dp)
        case("WT")
          WT = real(value,kind=dp)
        case("ymup")
          ymup = real(value,kind=dp)
        case("MZ")
          MZ = real(value,kind=dp)
        case("WC")
          WC = real(value,kind=dp)
        case("aS")
          aS = real(value,kind=dp)
        case("ymtau")
          ymtau = real(value,kind=dp)
        case("WH")
          WH = real(value,kind=dp)
        case("MC")
          MC = real(value,kind=dp)
        case("MU")
          MU = real(value,kind=dp)
        case("WB")
          WB = real(value,kind=dp)
        case("ymb")
          ymb = real(value,kind=dp)
        case("MD")
          MD = real(value,kind=dp)
        case("MS")
          MS = real(value,kind=dp)
        case default
        call warning_mdl('Trying to change unknown/internal parameter:' // trim&
        (param), where='set_parameter_mdl')
      end select 
    end subroutine set_parameter_mdl

    function get_parameter_mdl(param)
      character(*), intent(in) :: param
      complex(dp) :: get_parameter_mdl
      select case(param)
        case("MB")
          get_parameter_mdl = MB
        case("MT")
          get_parameter_mdl = MT
        case("yme")
          get_parameter_mdl = yme
        case("WW")
          get_parameter_mdl = WW
        case("ymt")
          get_parameter_mdl = ymt
        case("MH")
          get_parameter_mdl = MH
        case("MW")
          get_parameter_mdl = MW
        case("ymdo")
          get_parameter_mdl = ymdo
        case("aEW")
          get_parameter_mdl = aEW
        case("ymc")
          get_parameter_mdl = ymc
        case("cgghfin")
          get_parameter_mdl = cgghfin
        case("muMS")
          get_parameter_mdl = muMS
        case("WZ")
          get_parameter_mdl = WZ
        case("yms")
          get_parameter_mdl = yms
        case("ymm")
          get_parameter_mdl = ymm
        case("WT")
          get_parameter_mdl = WT
        case("ymup")
          get_parameter_mdl = ymup
        case("MZ")
          get_parameter_mdl = MZ
        case("WC")
          get_parameter_mdl = WC
        case("aS")
          get_parameter_mdl = aS
        case("ymtau")
          get_parameter_mdl = ymtau
        case("WH")
          get_parameter_mdl = WH
        case("MC")
          get_parameter_mdl = MC
        case("MU")
          get_parameter_mdl = MU
        case("WB")
          get_parameter_mdl = WB
        case("ymb")
          get_parameter_mdl = ymb
        case("MD")
          get_parameter_mdl = MD
        case("MS")
          get_parameter_mdl = MS
        case("sw2")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = sw2
        case("ZERO")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = ZERO
        case("yup")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = yup
        case("gw")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = gw
        case("ym")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = ym
        case("sw")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = sw
        case("ys")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = ys
        case("ee")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = ee
        case("g1")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = g1
        case("yt")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = yt
        case("ye")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = ye
        case("cggh")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = cggh
        case("vev")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = vev
        case("ytau")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = ytau
        case("cw")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = cw
        case("yc")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = yc
        case("lam")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = lam
        case("yb")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = yb
        case("ydo")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = ydo
        case("muH2")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = muH2
        case("gs")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = gs
        case default
          call warning_mdl('Trying to get an unknown parameter: ' // trim(param&
          ), where='get_parameter_mdl')
          get_parameter_mdl = cnul
      end select 
    end function get_parameter_mdl

    subroutine print_parameters_mdl(level)
      integer, intent(in) :: level
      character(len=15) :: fmts = '(4x,a10,g21.14)'
      character(len=24) :: fmtsc = '(4x,a10,g21.14,x,g21.14)'
      character(len=15) :: fmtm = '(4x,a7,es23.16)'
      character(len=25) :: fmtc = '(4x,a7,es23.16,x,es23.16)'
      character(len=26) :: fmtmw = '(4x,a7,es23.16,a8,es23.16)'
      character(len=18) :: fmtml = '(4x,a7,es23.16,a8)'
      write(nx,'(1x,75("-"))')
      write(nx,*) 'Input: Scale values:'
      write(nx,fmts) adjustl('muMS ='),real(muMS)
      write(nx,'(1x,75("-"))')
      write(nx,*) 'Input: Coupling values:'
      write(nx,fmtm) adjustl('aEW ='),real(aEW)
      write(nx,fmtm) adjustl('aS ='),real(aS)
      write(nx,fmtm) adjustl('cgghfin ='),real(cgghfin)
      write(nx,fmtm) adjustl('yme ='),real(yme)
      write(nx,fmtm) adjustl('ymm ='),real(ymm)
      write(nx,fmtm) adjustl('ymtau ='),real(ymtau)
      write(nx,'(1x,75("-"))')
      write(nx,*) 'Input: Pole masses and widths [GeV]:'
      write(nx,fmtmw) adjustl('MH ='),MH, adjustl('WH ='),WH
      if(MS_reg .eq. 2) then
        write(nx,fmtml) adjustl('MS ='), MS, 'light'
      else
        write(nx,fmtm) adjustl('MS ='), MS
      endif
      if(MU_reg .eq. 2) then
        write(nx,fmtml) adjustl('MU ='), MU, 'light'
      else
        write(nx,fmtm) adjustl('MU ='), MU
      endif
      if(MC_reg .eq. 2) then
        write(nx,fmtml) adjustl('MC ='), MC, 'light'
      else
        write(nx,fmtmw) adjustl('MC ='),MC, adjustl('WC ='),WC
      endif
      write(nx,fmtmw) adjustl('MT ='),MT, adjustl('WT ='),WT
      if(MD_reg .eq. 2) then
        write(nx,fmtml) adjustl('MD ='), MD, 'light'
      else
        write(nx,fmtm) adjustl('MD ='), MD
      endif
      if(MB_reg .eq. 2) then
        write(nx,fmtml) adjustl('MB ='), MB, 'light'
      else
        write(nx,fmtmw) adjustl('MB ='),MB, adjustl('WB ='),WB
      endif
      write(nx,fmtmw) adjustl('MW ='),MW, adjustl('WW ='),WW
      write(nx,fmtmw) adjustl('MZ ='),MZ, adjustl('WZ ='),WZ
      write(nx,'(1x,75("-"))')
      if (level .gt. 0) then
        write(nx,*) 'Complex masses squared [GeV^2] (internal):'
        write(nx,fmtc) adjustl('cMH2 ='), cMH2
        write(nx,fmtc) adjustl('cMS2 ='), cMS2
        write(nx,fmtc) adjustl('cMU2 ='), cMU2
        write(nx,fmtc) adjustl('cMC2 ='), cMC2
        write(nx,fmtc) adjustl('cMT2 ='), cMT2
        write(nx,fmtc) adjustl('cMD2 ='), cMD2
        write(nx,fmtc) adjustl('cMB2 ='), cMB2
        write(nx,fmtc) adjustl('cMW2 ='), cMW2
        write(nx,fmtc) adjustl('cMZ2 ='), cMZ2
        write(nx,'(1x,75("-"))')
        write(nx,*) 'Yukawa parameter values (internal):'
        write(nx,fmtc) adjustl('ymb ='),ymb
        write(nx,fmtc) adjustl('ymc ='),ymc
        write(nx,fmtc) adjustl('ymdo ='),ymdo
        write(nx,fmtc) adjustl('yms ='),yms
        write(nx,fmtc) adjustl('ymt ='),ymt
        write(nx,fmtc) adjustl('ymup ='),ymup
        write(nx,'(1x,75("-"))')
        write(nx,*) 'Internal parameter values:'
        write(nx,fmtc) 'ZERO =',ZERO
        write(nx,fmtc) 'cggh =',cggh
        write(nx,fmtc) 'cw =',cw
        write(nx,fmtc) 'ee =',ee
        write(nx,fmtc) 'g1 =',g1
        write(nx,fmtc) 'gs =',gs
        write(nx,fmtc) 'gw =',gw
        write(nx,fmtc) 'lam =',lam
        write(nx,fmtc) 'muH2 =',muH2
        write(nx,fmtc) 'sw =',sw
        write(nx,fmtc) 'sw2 =',sw2
        write(nx,fmtc) 'vev =',vev
        write(nx,fmtc) 'yb =',yb
        write(nx,fmtc) 'yc =',yc
        write(nx,fmtc) 'ydo =',ydo
        write(nx,fmtc) 'ye =',ye
        write(nx,fmtc) 'ym =',ym
        write(nx,fmtc) 'ys =',ys
        write(nx,fmtc) 'yt =',yt
        write(nx,fmtc) 'ytau =',ytau
        write(nx,fmtc) 'yup =',yup
        write(nx,'(1x,75("-"))')
      end if 
    end subroutine print_parameters_mdl

    subroutine set_resonant_particle_value_mdl(p,val)
      character(*), intent(in) :: p
      logical, intent(in) :: val
      if (particle_model_init) then
        call warning_mdl                                                       &
        ('Resonances need to be set before defining processes.', where         &
        ='set_resonant_particle_value_mdl')
      end if 
      select case(p)
        case("Z")
          Z_res = val
        case("W+")
          W_res = val
        case("W-")
          W_res = val
        case("c")
          c_res = val
        case("c~")
          c_res = val
        case("t")
          t_res = val
        case("t~")
          t_res = val
        case("b")
          b_res = val
        case("b~")
          b_res = val
        case("H")
          H_res = val
      case default
        call warning_mdl                                                       &
        ('Trying to set resonance for non-supported particle: ' // trim(p),    &
        where='set_resonant_particle_value_mdl')
      end select 
    end subroutine set_resonant_particle_value_mdl

    function is_resonant_particle_id_mdl(p) result(is_resonant_particle)
      integer, intent(in) :: p
      logical :: is_resonant_particle
      select case(p)
      case(P_Z)
        is_resonant_particle = Z_res
      case(P_WPlus)
        is_resonant_particle = W_res
      case(P_WMinus)
        is_resonant_particle = W_res
      case(P_c)
        is_resonant_particle = c_res
      case(P_cAnti)
        is_resonant_particle = c_res
      case(P_t)
        is_resonant_particle = t_res
      case(P_tAnti)
        is_resonant_particle = t_res
      case(P_b)
        is_resonant_particle = b_res
      case(P_bAnti)
        is_resonant_particle = b_res
      case(P_H)
        is_resonant_particle = H_res
      case default
        call warning_mdl                                                       &
        ('Trying to determine resonance status for non-support particle id.',  &
        where='is_resonant_particle_id_mdl')
        is_resonant_particle = .false.
      end select 
    end function is_resonant_particle_id_mdl

    subroutine set_light_particle_value_mdl(p,val)
      use input_mdl, only: warning_mdl
      character(*), intent(in) :: p
      logical, intent(in) :: val
      if (particle_model_init) then
        call warning_mdl                                                       &
        ('Light particles need to be set before defining processes.', where    &
        ='set_light_particle_value_mdl')
      end if 
      select case(p)
      case("u")
        MU_reg_light = val
      case("u~")
        MU_reg_light = val
      case("c")
        MC_reg_light = val
      case("c~")
        MC_reg_light = val
      case("d")
        MD_reg_light = val
      case("d~")
        MD_reg_light = val
      case("s")
        MS_reg_light = val
      case("s~")
        MS_reg_light = val
      case("b")
        MB_reg_light = val
      case("b~")
        MB_reg_light = val
      case default
        call warning_mdl                                                       &
        ('Trying to set light particle for non-supported particle: ' // trim(p &
        ), where='set_light_particle_value_mdl')
      end select 
    end subroutine set_light_particle_value_mdl

    function is_light_mass_mdl(m)
      use input_mdl, only: warning_mdl
      character(*), intent(in) :: m
      logical :: is_light_mass_mdl
      select case(m)
      case("MB_reg")
        is_light_mass_mdl = MB_reg_light
      case("MU_reg")
        is_light_mass_mdl = MU_reg_light
      case("MC_reg")
        is_light_mass_mdl = MC_reg_light
      case("MD_reg")
        is_light_mass_mdl = MD_reg_light
      case("MS_reg")
        is_light_mass_mdl = MS_reg_light
      case default
        call warning_mdl                                                       &
        ('Trying to determine light status for non-support mass: ' // trim(m), &
        where='is_light_mass_mdl')
        is_light_mass_mdl = .false.
      end select 
    end function is_light_mass_mdl

    function is_particle_mdl(p) result(is_particle)
      integer, intent(in) :: p
      logical :: is_particle
      character(len=3) :: p_str
      select case(p)
      case(P_A)
        is_particle = .true.
      case(P_Z)
        is_particle = .true.
      case(P_WPlus)
        is_particle = .false.
      case(P_WMinus)
        is_particle = .true.
      case(P_g)
        is_particle = .true.
      case(P_ghG)
        is_particle = .true.
      case(P_ghGAnti)
        is_particle = .false.
      case(P_u)
        is_particle = .true.
      case(P_uAnti)
        is_particle = .false.
      case(P_c)
        is_particle = .true.
      case(P_cAnti)
        is_particle = .false.
      case(P_t)
        is_particle = .true.
      case(P_tAnti)
        is_particle = .false.
      case(P_d)
        is_particle = .true.
      case(P_dAnti)
        is_particle = .false.
      case(P_s)
        is_particle = .true.
      case(P_sAnti)
        is_particle = .false.
      case(P_b)
        is_particle = .true.
      case(P_bAnti)
        is_particle = .false.
      case(P_H)
        is_particle = .true.
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="is_particle_mdl")
      end select 
    end function is_particle_mdl

    function is_backgroundfield_mdl(p) result(isbf)
      integer, intent(in) :: p
      logical :: isbf
      character(len=3) :: p_str
      select case(p)
      case(P_A)
        isbf = .true.
      case(P_Z)
        isbf = .true.
      case(P_WPlus)
        isbf = .true.
      case(P_WMinus)
        isbf = .true.
      case(P_g)
        isbf = .true.
      case(P_ghG)
        isbf = .true.
      case(P_ghGAnti)
        isbf = .true.
      case(P_u)
        isbf = .true.
      case(P_uAnti)
        isbf = .true.
      case(P_c)
        isbf = .true.
      case(P_cAnti)
        isbf = .true.
      case(P_t)
        isbf = .true.
      case(P_tAnti)
        isbf = .true.
      case(P_d)
        isbf = .true.
      case(P_dAnti)
        isbf = .true.
      case(P_s)
        isbf = .true.
      case(P_sAnti)
        isbf = .true.
      case(P_b)
        isbf = .true.
      case(P_bAnti)
        isbf = .true.
      case(P_H)
        isbf = .true.
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="is_backgroundfield_mdl")
      end select 
    end function is_backgroundfield_mdl

    function is_quantumfield_mdl(p) result(isqf)
      integer, intent(in) :: p
      logical :: isqf
      character(len=3) :: p_str
      select case(p)
      case(P_A)
        isqf = .true.
      case(P_Z)
        isqf = .true.
      case(P_WPlus)
        isqf = .true.
      case(P_WMinus)
        isqf = .true.
      case(P_g)
        isqf = .true.
      case(P_ghG)
        isqf = .true.
      case(P_ghGAnti)
        isqf = .true.
      case(P_u)
        isqf = .true.
      case(P_uAnti)
        isqf = .true.
      case(P_c)
        isqf = .true.
      case(P_cAnti)
        isqf = .true.
      case(P_t)
        isqf = .true.
      case(P_tAnti)
        isqf = .true.
      case(P_d)
        isqf = .true.
      case(P_dAnti)
        isqf = .true.
      case(P_s)
        isqf = .true.
      case(P_sAnti)
        isqf = .true.
      case(P_b)
        isqf = .true.
      case(P_bAnti)
        isqf = .true.
      case(P_H)
        isqf = .true.
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="is_quantumfield_mdl")
      end select 
    end function is_quantumfield_mdl

    function has_feature_mdl(feature) result(has_feature)
      character(*), intent(in) :: feature
      logical :: has_feature
      select case(feature)
      case("fermionloop_opt")
        has_feature = .true.
      case("qcd_rescaling")
        has_feature = .false.
      case default
        has_feature = .false.
      end select 
    end function has_feature_mdl

    function get_particle_type_mdl(particle_in) result(particle_type)
      integer, intent(in) :: particle_in
      character(2) :: particle_type
      select case (get_particle_spin_mdl(particle_in))
      ! ghost
      case(-1)
        select case (is_particle_mdl(particle_in))
        ! ghost
        case (.true.)
          particle_type = 'x'
        ! antighost
        case (.false.)
          particle_type = 'x~'
        end select
      ! scalar
      case (1)
        particle_type = 's'
      ! fermion
      case (2)
        select case (is_particle_mdl(particle_in))
        ! fermion
        case (.true.)
          particle_type = 'f'
        ! antifermion
        case (.false.)
          particle_type = 'f~'
        end select
      ! vector
      case (3)
        particle_type = 'v'
      case default
        particle_type = '0'
      end select
    end function get_particle_type_mdl

    function get_particle_type2_mdl(particle_in) result(particle_type)
      integer, intent(in) :: particle_in
      character(2) :: particle_type
      select case (get_particle_spin_mdl(particle_in))
      ! ghost
      case(-1)
        select case (is_particle_mdl(particle_in))
        ! ghost
        case (.true.)
          select case (get_particle_colour_mdl(particle_in) .ne. 1)
          ! colored ghost
          case (.true.)
            particle_type = 'G'
          ! ghost
          case (.false.)
            particle_type = 'x'
          end select
        ! antighost
        case (.false.)
          select case (get_particle_colour_mdl(particle_in) .ne. 1)
          ! colored antighost
          case (.true.)
            particle_type = 'G~'
          ! antighost
          case (.false.)
            particle_type = 'x~'
          end select
        end select
      ! scalar
      case (1)
        particle_type = 's'
      ! fermion
      case (2)
        select case (is_particle_mdl(particle_in))
        ! fermion
        case (.true.)
          select case (get_particle_colour_mdl(particle_in) .ne. 1)
          ! quark
          case (.true.)
            particle_type = 'q'
          ! fermion
          case (.false.)
            particle_type = 'f'
          end select
        ! antifermion
        case (.false.)
          select case (get_particle_colour_mdl(particle_in) .ne. 1)
          ! antiquark
          case (.true.)
            particle_type = 'q~'
          ! antifermion
          case (.false.)
            particle_type = 'f~'
          end select
        end select
      ! vector
      case (3)
        select case (get_particle_colour_mdl(particle_in) .ne. 1)
        ! gluon
        case (.true.)
          particle_type = 'g'
        ! vector
        case (.false.)
          particle_type = 'v'
        end select
      case default
        particle_type = '0'
      end select
    end function get_particle_type2_mdl

    function get_axodraw_propagator_mdl(particle_type) result(axoprop)
      character(2), intent(in) :: particle_type
      character(20) :: axoprop
      select case (particle_type)
        case ('g')
          axoprop = '\Gluon'
        case ('v')
          axoprop = '\Photon'
        case ('f', 'f~')
          axoprop = '\ArrowLine'
        case ('q', 'q~')
          axoprop = '\ArrowLine'
        case ('x', 'x~')
          axoprop = '\DashArrowLine'
        case ('G', 'G~')
          axoprop = '\DashLine'
        case ('s')
          axoprop = '\DashLine'
        case default
          write(*,*) "particle_type:", particle_type
          write (*, *) 'Case not implemented in axodraw_propagator'
          stop
      end select
    end function get_axodraw_propagator_mdl
end module class_particles
