!******************************************************************************!
!                                                                              !
!    class_vertices.f90                                                        !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016-2018 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!
!                                                           .---.              !
!                                                          /  .  \             !
!                                                         |\_/|   |            !
!                                                         |   |  /|            !
!          .----------------------------------------------------' |            !
!         /  .-.   class_vertices.f90                             |            !
!        |  /   \  This code was generated with REPT1L.           |            !
!        | |\_.  | Change this file and you get what you deserve. |            !
!        |\|  | /| Author: J.N. Lang                              |            !
!        | `---' | UFO date: 03/2019                              |            !
!        |       | UFO version: 1.0                               |            !
!        |       | Generation date: 2019-10-14 13:42             /             !
!        |       |----------------------------------------------'              !
!        \       |                                                             !
!         \     /                                                              !
!          `---'                                                               !
!******************************************************************************!
module class_vertices
  use class_particles, only: n_orders
  use constants_mdl
  use dictionary, only: dict_init
  implicit none
  save

  type Branch
    integer :: lorentz_id
    integer :: perm
    integer, allocatable :: couplings(:)
    integer, dimension(n_orders) :: coupling_order
    integer :: color_id
    integer :: colorfac_id
    ! Propagator extension (only supported for massive vectors)
    ! 0: Usual propagator
    ! 1: Rxi-propagator part with pole @ M_V^2
    ! 2: Rxi-propagator part with pole @ \xi*M_V^2
    integer :: pext
  end type

  type Vertex
    type(Branch), allocatable :: branch(:)
    type(Branch), allocatable :: branchCT(:)
    type(Branch), allocatable :: branchR2(:)
  end type

  logical :: vertices_init = .false.
  logical :: vertices_filled = .false.
  integer, parameter :: n_vertices = 669
  type(Vertex) :: vertices(0:n_vertices-1)

  ! Defines the maximum allowed number of legs participating at a vertex
  integer, parameter :: max_leg_multiplicity = 7

  integer, parameter :: FFaS0 = 1
  integer, parameter :: VSFSFa0 = 2
  integer, parameter :: VVSSVVS0 = 3
  integer, parameter :: SVVVS0 = 4
  integer, parameter :: SFaVFS0 = 5
  integer, parameter :: FFaV0 = 6
  integer, parameter :: SVVVV0 = 7
  integer, parameter :: SVVVV1 = 8
  integer, parameter :: FaSFSV0 = 9
  integer, parameter :: SVSVSVV0 = 10
  integer, parameter :: VVSVSS0 = 11
  integer, parameter :: FFaVS0 = 12
  integer, parameter :: VVSVSV0 = 13
  integer, parameter :: SVVVSVS0 = 14
  integer, parameter :: SSVVV0 = 15
  integer, parameter :: VVSS0 = 16
  integer, parameter :: FaSVF0 = 17
  integer, parameter :: SSVVS0 = 18
  integer, parameter :: VVSV0 = 19
  integer, parameter :: SVVSVVS0 = 20
  integer, parameter :: VVV0 = 21
  integer, parameter :: SVVSSVV0 = 22
  integer, parameter :: VVS0 = 23
  integer, parameter :: VVS1 = 24
  integer, parameter :: VSVVSVS0 = 25
  integer, parameter :: VSFFa0 = 26
  integer, parameter :: FFa0 = 27
  integer, parameter :: VSVVV0 = 28
  integer, parameter :: VSVVV1 = 29
  integer, parameter :: FaFSV0 = 30
  integer, parameter :: FVFaSS0 = 31
  integer, parameter :: FaFSS0 = 32
  integer, parameter :: VUaU0 = 33
  integer, parameter :: VFSFa0 = 34
  integer, parameter :: FSFaVS0 = 35
  integer, parameter :: SVSSVVV0 = 36
  integer, parameter :: SVSVVS0 = 37
  integer, parameter :: SVSVVV0 = 38
  integer, parameter :: FaFVS0 = 39
  integer, parameter :: FaFS0 = 40
  integer, parameter :: FSVFaS0 = 41
  integer, parameter :: VSVSS0 = 42
  integer, parameter :: VSVSVSV0 = 43
  integer, parameter :: FaFV0 = 44
  integer, parameter :: SSSS0 = 45
  integer, parameter :: SFSVFa0 = 46
  integer, parameter :: FaSSVF0 = 47
  integer, parameter :: SFFa0 = 48
  integer, parameter :: SFaF0 = 49
  integer, parameter :: SVFSFa0 = 50
  integer, parameter :: SVV0 = 51
  integer, parameter :: SVV1 = 52
  integer, parameter :: FSFaV0 = 53
  integer, parameter :: FSFaS0 = 54
  integer, parameter :: SVSVVSV0 = 55
  integer, parameter :: VSSSVVV0 = 56
  integer, parameter :: VSFaFS0 = 57
  integer, parameter :: FSSVFa0 = 58
  integer, parameter :: VVSVSVS0 = 59
  integer, parameter :: FFaVSS0 = 60
  integer, parameter :: SVFFa0 = 61
  integer, parameter :: VVVSSS0 = 62
  integer, parameter :: VVVSSV0 = 63
  integer, parameter :: VSSFFa0 = 64
  integer, parameter :: FaFSSV0 = 65
  integer, parameter :: SVFFaS0 = 66
  integer, parameter :: VSSVVSV0 = 67
  integer, parameter :: VSSSVV0 = 68
  integer, parameter :: FaVSFS0 = 69
  integer, parameter :: FSFa0 = 70
  integer, parameter :: SSVVVSV0 = 71
  integer, parameter :: FVSFaS0 = 72
  integer, parameter :: VVSVVSS0 = 73
  integer, parameter :: VSSVVVS0 = 74
  integer, parameter :: VVSSSV0 = 75
  integer, parameter :: SFaFSV0 = 76
  integer, parameter :: SFVFaS0 = 77
  integer, parameter :: SSFVFa0 = 78
  integer, parameter :: SVSVVVS0 = 79
  integer, parameter :: VFaFS0 = 80
  integer, parameter :: FSSFaV0 = 81
  integer, parameter :: SVVVVS0 = 82
  integer, parameter :: SVVVVSS0 = 83
  integer, parameter :: FaSVSF0 = 84
  integer, parameter :: SFFaV0 = 85
  integer, parameter :: SFFaS0 = 86
  integer, parameter :: VVVS0 = 87
  integer, parameter :: VSFFaS0 = 88
  integer, parameter :: VVVV0 = 89
  integer, parameter :: SFFaSV0 = 90
  integer, parameter :: SVSV0 = 91
  integer, parameter :: VVSSSVV0 = 92
  integer, parameter :: VSVS0 = 93
  integer, parameter :: VFFaSS0 = 94
  integer, parameter :: VSVV0 = 95
  integer, parameter :: SFSFa0 = 96
  integer, parameter :: VFaSFS0 = 97
  integer, parameter :: SFaVSF0 = 98
  integer, parameter :: SVFaFS0 = 99
  integer, parameter :: SVSSV0 = 100
  integer, parameter :: VSSVVS0 = 101
  integer, parameter :: VSSSV0 = 102
  integer, parameter :: VSSVVV0 = 103
  integer, parameter :: SSFFaV0 = 104
  integer, parameter :: FVSSFa0 = 105
  integer, parameter :: FSVFa0 = 106
  integer, parameter :: SSVVSV0 = 107
  integer, parameter :: FaSF0 = 108
  integer, parameter :: FaSFS0 = 109
  integer, parameter :: SFaFV0 = 110
  integer, parameter :: SVFaSF0 = 111
  integer, parameter :: FFaSV0 = 112
  integer, parameter :: FaVSSF0 = 113
  integer, parameter :: SSVV0 = 114
  integer, parameter :: VFFaS0 = 115
  integer, parameter :: SSVSV0 = 116
  integer, parameter :: SSS0 = 117
  integer, parameter :: SFVFa0 = 118
  integer, parameter :: VSVSVS0 = 119
  integer, parameter :: FVSFa0 = 120
  integer, parameter :: VSVSVV0 = 121
  integer, parameter :: VSVVSV0 = 122
  integer, parameter :: VSSFaF0 = 123
  integer, parameter :: VFFa0 = 124
  integer, parameter :: VSVVSS0 = 125
  integer, parameter :: UaUV0 = 126
  integer, parameter :: SSVVVS0 = 127
  integer, parameter :: SVFaF0 = 128
  integer, parameter :: SSVVVV0 = 129
  integer, parameter :: SFaVF0 = 130
  integer, parameter :: VVVSS0 = 131
  integer, parameter :: SVVV0 = 132
  integer, parameter :: VVVSV0 = 133
  integer, parameter :: VVVSV1 = 134
  integer, parameter :: SSVSVV0 = 135
  integer, parameter :: FSSFa0 = 136
  integer, parameter :: VSVSVVS0 = 137
  integer, parameter :: FSFaSV0 = 138
  integer, parameter :: SFaFVS0 = 139
  integer, parameter :: SFaFS0 = 140
  integer, parameter :: FaSFV0 = 141
  integer, parameter :: SVVS0 = 142
  integer, parameter :: VVVSVS0 = 143
  integer, parameter :: SFFaVS0 = 144
  integer, parameter :: FVFaS0 = 145
  integer, parameter :: VVSVSSV0 = 146
  integer, parameter :: VSSVV0 = 147
  integer, parameter :: SVSVS0 = 148
  integer, parameter :: VVVSVSS0 = 149
  integer, parameter :: SVSVSV0 = 150
  integer, parameter :: FFaSVS0 = 151
  integer, parameter :: SVSVV0 = 152
  integer, parameter :: VSSVS0 = 153
  integer, parameter :: FVFa0 = 154
  integer, parameter :: FaF0 = 155
  integer, parameter :: SSSVVV0 = 156
  integer, parameter :: VVSVV0 = 157
  integer, parameter :: VVSVV1 = 158
  integer, parameter :: VFSFaS0 = 159
  integer, parameter :: VVSVS0 = 160
  integer, parameter :: SFaSVF0 = 161
  integer, parameter :: SVVSS0 = 162
  integer, parameter :: SSVFaF0 = 163
  integer, parameter :: VUUa0 = 164
  integer, parameter :: FFaSS0 = 165
  integer, parameter :: FaSVFS0 = 166
  integer, parameter :: VVVSSSV0 = 167
  integer, parameter :: SVSFaF0 = 168
  integer, parameter :: SSSVV0 = 169
  integer, parameter :: VVVVSSS0 = 170
  integer, parameter :: FaVFSS0 = 171
  integer, parameter :: VSVSV0 = 172
  integer, parameter :: SFaSFV0 = 173
  integer, parameter :: SVVSSV0 = 174
  integer, parameter :: VSFaSF0 = 175
  integer, parameter :: VSFaF0 = 176
  integer, parameter :: VSSVSV0 = 177
  integer, parameter :: VFaF0 = 178
  integer, parameter :: SSVVSVV0 = 179
  integer, parameter :: VVSSV0 = 180
  integer, parameter :: VVSSS0 = 181
  integer, parameter :: FFaSSV0 = 182
  integer, parameter :: UUaV0 = 183
  integer, parameter :: SVSFFa0 = 184
  integer, parameter :: VVVVSS0 = 185
  integer, parameter :: SFaSF0 = 186
  integer, parameter :: FaVFS0 = 187
  integer, parameter :: SSFaF0 = 188
  integer, parameter :: VSSVSVV0 = 189
  integer, parameter :: SVVSVV0 = 190
  integer, parameter :: FSVSFa0 = 191
  integer, parameter :: VSVVVS0 = 192
  integer, parameter :: SVVSVS0 = 193
  integer, parameter :: SFSFaV0 = 194
  integer, parameter :: SSFaVF0 = 195
  integer, parameter :: SSVSVVV0 = 196
  integer, parameter :: SSFaFV0 = 197
  integer, parameter :: VSVVS0 = 198
  integer, parameter :: VVVVS0 = 199
  integer, parameter :: VVVVS1 = 200
  integer, parameter :: FaVF0 = 201
  integer, parameter :: SVVVSSV0 = 202
  integer, parameter :: FaSSFV0 = 203
  integer, parameter :: UaVU0 = 204
  integer, parameter :: VSVVVSS0 = 205
  integer, parameter :: VVSSVSV0 = 206
  integer, parameter :: SSSVVVV0 = 207
  integer, parameter :: VVSVVS0 = 208
  integer, parameter :: SVVVSV0 = 209
  integer, parameter :: FaFVSS0 = 210
  integer, parameter :: SVVVSS0 = 211
  integer, parameter :: SFVSFa0 = 212
  integer, parameter :: VFaSSF0 = 213
  integer, parameter :: VSVSSVV0 = 214
  integer, parameter :: SSVVVVS0 = 215
  integer, parameter :: SSVFFa0 = 216
  integer, parameter :: VV0 = 217
  integer, parameter :: VV1 = 218
  integer, parameter :: FaVSF0 = 219
  integer, parameter :: FaSFVS0 = 220
  integer, parameter :: VSVSSV0 = 221
  integer, parameter :: VSVVSSV0 = 222
  integer, parameter :: FaFSVS0 = 223
  integer, parameter :: VFSSFa0 = 224
  integer, parameter :: FaSSF0 = 225
  integer, parameter :: VFaFSS0 = 226
  integer, parameter :: VSV0 = 227
  integer, parameter :: VSV1 = 228
  integer, parameter :: VVVSSVS0 = 229
  integer, parameter :: SSFFa0 = 230
  integer, parameter :: UVUa0 = 231
  integer, parameter :: VVSSVS0 = 232
  integer, parameter :: VSSV0 = 233
  integer, parameter :: VVSSVV0 = 234
  integer, parameter :: SVSSVV0 = 235
  integer, parameter :: SVVSVSV0 = 236
  integer, parameter :: VFaSF0 = 237
  integer, parameter :: SVVSV0 = 238
  complex(dp), parameter :: c1 = cmplx(-1d0/2d0, 0d0, kind=dp)
  character(len=20), parameter :: c1f = "-1/2"
  complex(dp), parameter :: c2 = cmplx(1d0/2d0, 0d0, kind=dp)
  character(len=20), parameter :: c2f = "1/2"
  complex(dp), parameter :: c3 = cmplx(1d0, 0d0, kind=dp)
  character(len=20), parameter :: c3f = "1"
  complex(dp), parameter :: c4 = cmplx(0d0, sq2/2d0, kind=dp)
  character(len=20), parameter :: c4f = "i_*sq2/2"
  complex(dp), parameter :: c5 = cmplx(0d0, -sq2/2d0, kind=dp)
  character(len=20), parameter :: c5f = "-i_*sq2/2"
  complex(dp), parameter :: c6 = cmplx(-1d0/Nc, 0d0, kind=dp)
  character(len=20), parameter :: c6f = "-1/Nc"
  complex(dp), parameter :: c7 = cmplx(sq2/2d0, 0d0, kind=dp)
  character(len=20), parameter :: c7f = "sq2/2"
  complex(dp), parameter :: c8 = cmplx(-sq2/(2d0*Nc), 0d0, kind=dp)
  character(len=20), parameter :: c8f = "-sq2/(2*Nc)"
  complex(dp), dimension(1:8) :: colorcoeffs = (/c1, c2, c3, c4, c5, c6, c7, c8&
                                               /)
  character(len=20), dimension(1:8) :: colorcoeffs_form = (/c1f, c2f, c3f, c4f &
                 , c5f, c6f, c7f, c8f/)
  integer, dimension(1:238) :: lorentz_rank = (/-1, -1, 0, 0, -1, 0, 0, -1, -1 &
                                              , 0, 1, -1, 0, 0, 1, 2, -1, 2, 1 &
                                              , 0, 1, 0, 2, -1, 0, -1, -1, 0, -&
                                              1, -1, -1, -1, 1, -1, -1, 0, 0, 0&
                                              , -1, -1, -1, 2, 0, 0, 0, -1, -1 &
                                              , -1, -1, -1, 2, -1, -1, -1, 0, 0&
                                              , -1, -1, 0, -1, -1, 1, 0, -1, -1&
                                              , -1, 0, 1, -1, -1, 0, -1, 0, 0, &
                                              1, -1, -1, -1, 0, -1, -1, 0, 0, -&
                                              1, -1, -1, 1, -1, 0, -1, 2, 0, 2 &
                                              , -1, 1, -1, -1, -1, -1, 2, 1, 2 &
                                              , 0, -1, -1, -1, 1, -1, -1, -1, -&
                                              1, -1, -1, 2, -1, 2, 0, -1, 1, -1&
                                              , 0, 0, -1, 1, 1, 1, 0, -1, 0, -1&
                                              , 1, 1, 0, -1, 1, -1, 0, -1, -1  &
                                              , -1, -1, 0, 0, -1, -1, 0, 1, 2, &
                                              0, 1, -1, 1, 2, 1, -1, 1, 0, -1  &
                                              , -1, 1, -1, 2, -1, 1, -1, -1, 0 &
                                              , -1, 2, 0, -1, 1, -1, 1, -1, -1 &
                                              , 1, 1, 0, 1, 2, -1, 1, -1, 0, -1&
                                              , -1, -1, 0, 0, -1, 0, 0, -1, -1 &
                                              , 0, -1, 1, 0, -1, 1, 0, -1, 1, 0&
                                              , 0, 0, 0, 0, -1, 0, -1, -1, 0, 0&
                                              , -1, -1, -1, -1, -1, 1, 0, -1, -&
                                              1, -1, -1, 2, -1, 0, -1, 1, 1, 2 &
                                              , 0, 1, 0, -1, 1/)
  integer, dimension(1:238) :: ff_ids = (/0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
                                        , 0, 0, 0, 0/)
  integer, dimension(1:238) :: recola_base = (/0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
                                             , 0, 0, 0, 0, 0, 0, 0/)
  integer, dimension(1:238) :: propagator_extension = (/0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0, 0, 0 &
                                                      , 0, 0, 0, 0, 0, 0/)
  character(len=*), parameter :: model_timestamp = "Mon Oct 14 13:28:14 2019"

  contains
    function get_timestamp_mdl() result(ret_timestamp)
      character(len=len(model_timestamp)) :: ret_timestamp
      ret_timestamp = model_timestamp
    end function get_timestamp_mdl

    function get_lorentz_rank_mdl(l_id) result(ret_lorentz_rank)
      integer, intent(in) :: l_id
      integer :: ret_lorentz_rank
      ret_lorentz_rank = lorentz_rank(l_id)
    end function get_lorentz_rank_mdl

    function get_propagator_extension_mdl(l_id) result(ret_pext)
      integer, intent(in) :: l_id
      integer :: ret_pext
      ret_pext = propagator_extension(l_id)
    end function get_propagator_extension_mdl

    function get_max_leg_multiplicity_mdl() result(ret_max_leg_multiplicity)
      integer :: ret_max_leg_multiplicity
      ret_max_leg_multiplicity = max_leg_multiplicity
    end function get_max_leg_multiplicity_mdl

    function get_fourfermion_id_mdl(l_id) result(ret_ffid)
      integer, intent(in) :: l_id
      integer :: ret_ffid
      ret_ffid = ff_ids(l_id)
    end function get_fourfermion_id_mdl

    function is_model_init_mdl() result(ret_init)
      logical :: ret_init
      ret_init = vertices_init
    end function is_model_init_mdl

    function is_model_filled_mdl() result(ret_init)
      logical :: ret_init
      ret_init = vertices_filled
    end function is_model_filled_mdl

    subroutine createCurrents
      use fill_currents, only: fill_currents_mdl
      implicit none
      call fill_currents_mdl(.false.)
    end subroutine createCurrents

    subroutine check_vertices_mdl
      use fill_currents, only: fill_currents_mdl
      implicit none
      call fill_currents_mdl(.true.)
    end subroutine check_vertices_mdl

    subroutine av(vid,bid,cid,cfid,pext,lid,perm,co,cop)
      integer, intent(in) :: vid,bid,cid,cfid,pext,lid,perm
      integer, dimension(0:), intent(in) :: co, cop
      integer scop,i
      vertices(vid)%branch(bid)%color_id = cid
      vertices(vid)%branch(bid)%colorfac_id = cfid
      vertices(vid)%branch(bid)%pext = pext
      vertices(vid)%branch(bid)%lorentz_id = lid
      vertices(vid)%branch(bid)%perm = perm
      vertices(vid)%branch(bid)%coupling_order = co
      scop = size(cop)
      allocate(vertices(vid)%branch(bid)%couplings(0:scop-1))
      do i = 0, scop-1
        vertices(vid)%branch(bid)%couplings(i) = cop(i)
      end do
    end subroutine av

    subroutine avCT(vid,bid,cid,cfid,pext,lid,perm,co,cop)
      integer, intent(in) :: vid,bid,cid,cfid,pext,lid,perm
      integer, dimension(0:), intent(in) :: co, cop
      integer scop,i
      vertices(vid)%branchCT(bid)%color_id = cid
      vertices(vid)%branchCT(bid)%colorfac_id = cfid
      vertices(vid)%branchCT(bid)%pext = pext
      vertices(vid)%branchCT(bid)%lorentz_id = lid
      vertices(vid)%branchCT(bid)%perm = perm
      vertices(vid)%branchCT(bid)%coupling_order = co
      scop = size(cop)
      allocate(vertices(vid)%branchCT(bid)%couplings(0:scop-1))
      do i = 0, scop-1
        vertices(vid)%branchCT(bid)%couplings(i) = cop(i)
      end do
    end subroutine avCT

    subroutine avR2(vid,bid,cid,cfid,pext,lid,perm,co,cop)
      integer, intent(in) :: vid,bid,cid,cfid,pext,lid,perm
      integer, dimension(0:), intent(in) :: co, cop
      integer scop,i
      vertices(vid)%branchR2(bid)%color_id = cid
      vertices(vid)%branchR2(bid)%colorfac_id = cfid
      vertices(vid)%branchR2(bid)%pext = pext
      vertices(vid)%branchR2(bid)%lorentz_id = lid
      vertices(vid)%branchR2(bid)%perm = perm
      vertices(vid)%branchR2(bid)%coupling_order = co
      scop = size(cop)
      allocate(vertices(vid)%branchR2(bid)%couplings(0:scop-1))
      do i = 0, scop-1
        vertices(vid)%branchR2(bid)%couplings(i) = cop(i)
      end do
    end subroutine avR2

    function get_colourfac_mdl(cfid)
      integer, intent(in) :: cfid
      complex(dp) :: get_colourfac_mdl
      get_colourfac_mdl = colorcoeffs(cfid)
    end function get_colourfac_mdl

    function get_colourfac_form_mdl(cfid)
      integer, intent(in) :: cfid
      character(len=20) :: get_colourfac_form_mdl
      get_colourfac_form_mdl = colorcoeffs_form(cfid)
    end function get_colourfac_form_mdl

  function get_recola_base_mdl(l_id) result(ret_rb)
      integer, intent(in) :: l_id
      integer :: ret_rb
      ret_rb = recola_base(l_id)
    end function get_recola_base_mdl

  subroutine get_vertex_branch_status_mdl(vx,xlp,branch_exists)
    integer, intent(in)  :: vx,xlp
    logical, intent(out) ::branch_exists
    select case (xlp)
    case(TreeBranch:LoopBranch)
      branch_exists = allocated(vertices(vx)%branch)
    case(CTBranch)
      branch_exists = allocated(vertices(vx)%branchCT)
    case(R2Branch)
      branch_exists = allocated(vertices(vx)%branchR2)
    end select
  end subroutine get_vertex_branch_status_mdl

  subroutine get_vertex_colour_branches_mdl(vx,xlp,cfids,nbranches)
    integer, allocatable, intent(out) :: cfids(:)
    integer, intent(out) :: nbranches
    integer, intent(in)  :: vx,xlp
    select case (xlp)
    case (TreeBranch:LoopBranch)
      allocate(cfids(size(vertices(vx)%branch)))
      cfids(:) = vertices(vx)%branch(0:)%color_id
      nbranches = size(vertices(vx)%branch)
    case(CTBranch)
      allocate(cfids(size(vertices(vx)%branchCT)))
      cfids(:) = vertices(vx)%branchCT(0:)%color_id
      nbranches = size(vertices(vx)%branchCT)
    case(R2Branch)
      allocate(cfids(size(vertices(vx)%branchR2)))
      cfids(:) = vertices(vx)%branchR2(0:)%color_id
      nbranches = size(vertices(vx)%branchR2)
    end select
  end subroutine get_vertex_colour_branches_mdl

  subroutine get_vertex_pext_mdl(vx,xlp,br,pext)
    integer, intent(in) :: vx,xlp,br
    integer, intent(out) :: pext
    select case (xlp)
    case (TreeBranch:LoopBranch)
      pext = vertices(vx)%branch(br)%pext
    case(CTBranch)
      pext = vertices(vx)%branchCT(br)%pext
    case(R2Branch)
      pext = vertices(vx)%branchR2(br)%pext
    end select
  end subroutine get_vertex_pext_mdl

  subroutine get_vertex_order_increase_mdl(vx,xlp,br,orderInc)
    integer, intent(in) :: vx,xlp,br
    integer, dimension(:), intent(out) :: orderInc
    select case (xlp)
    case (TreeBranch:LoopBranch)
      orderInc = vertices(vx)%branch(br)%coupling_order
    case(CTBranch)
      orderInc = vertices(vx)%branchCT(br)%coupling_order
    case(R2Branch)
      orderInc = vertices(vx)%branchR2(br)%coupling_order
    end select
  end subroutine get_vertex_order_increase_mdl

  subroutine get_vertex_lorentz_type_mdl(vx,xlp,br,lid)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: lid
    select case (xlp)
    case (TreeBranch:LoopBranch)
      lid = vertices(vx)%branch(br)%lorentz_id
    case(CTBranch)
      lid = vertices(vx)%branchCT(br)%lorentz_id
    case(R2Branch)
      lid = vertices(vx)%branchR2(br)%lorentz_id
    end select
  end subroutine get_vertex_lorentz_type_mdl

  subroutine get_vertex_perm_mdl(vx,xlp,br,perm)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: perm
    select case (xlp)
    case (TreeBranch:LoopBranch)
      perm = vertices(vx)%branch(br)%perm
    case(CTBranch)
      perm = vertices(vx)%branchCT(br)%perm
    case(R2Branch)
      perm = vertices(vx)%branchR2(br)%perm
    end select
  end subroutine get_vertex_perm_mdl

  subroutine get_vertex_colour_id_mdl(vx,xlp,br,cid)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: cid
    select case (xlp)
    case (TreeBranch:LoopBranch)
      cid = vertices(vx)%branch(br)%colorfac_id
    case(CTBranch)
      cid = vertices(vx)%branchCT(br)%colorfac_id
    case(R2Branch)
      cid = vertices(vx)%branchR2(br)%colorfac_id
    end select
  end subroutine get_vertex_colour_id_mdl

  function is_zero_coupling_vertex_mdl(vx,xlp,br) result(zero)
    use class_couplings, only: get_coupling_value_mdl
    use input_mdl, only: zerocut
    integer, intent(in)  :: vx,xlp,br
    logical              :: zero
    real(dp)             :: maxvalue
    integer              :: i

    zero = .false.
    select case (xlp)
    case (TreeBranch:LoopBranch)
      maxvalue = 0d0
      do i = 1,size(vertices(vx)%branch(br)%couplings)
        maxvalue = maxvalue + &
                abs(get_coupling_value_mdl(vertices(vx)%branch(br)%couplings(i-&
                1)))
      end do
      if (maxvalue/zerocut .lt. 100d0) then
        zero = .true.
      end if
    case(CTBranch)
      maxvalue = 0d0
      do i = 1,size(vertices(vx)%branchCT(br)%couplings)
        maxvalue = maxvalue + &
                abs(get_coupling_value_mdl(vertices(vx)%branchCT(br)%couplings(&
                i-1)))
      end do
      if (maxvalue/zerocut .lt. 100d0) then
        zero = .true.
      end if
    case(R2Branch)
      maxvalue = 0d0
      do i = 1,size(vertices(vx)%branchR2(br)%couplings)
        maxvalue = maxvalue + &
                abs(get_coupling_value_mdl(vertices(vx)%branchR2(br)%couplings(&
                i-1)))
      end do
      if (maxvalue/zerocut .lt. 100d0) then
        zero = .true.
      end if
    end select
  end function is_zero_coupling_vertex_mdl

  subroutine get_vertex_ncouplings_mdl(vx,xlp,br,ncouplings)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: ncouplings
    select case(xlp)
    case(TreeBranch, LoopBranch)
      ncouplings = size(vertices(vx)%branch(br)%couplings)
    case (CTBranch)
      ncouplings = size(vertices(vx)%branchCT(br)%couplings)
    case (R2Branch)
      ncouplings = size(vertices(vx)%branchR2(br)%couplings)
    end select
  end subroutine get_vertex_ncouplings_mdl

  subroutine get_vertex_coupling_ids_mdl(vx,xlp,br,cids)
    integer, intent(in)  :: vx,xlp,br
    integer, intent(out) :: cids(:)
    select case(xlp)
    case(TreeBranch, LoopBranch)
      cids(:) = vertices(vx)%branch(br)%couplings(:)
    case (CTBranch)
      cids(:) = vertices(vx)%branchCT(br)%couplings(:)
    case (R2Branch)
      cids(:) = vertices(vx)%branchR2(br)%couplings(:)
    end select
  end subroutine get_vertex_coupling_ids_mdl

  subroutine init_vertices_mdl()
    use input_mdl, only: nx, error_mdl
    use class_particles,only: n_orders,driver_timestamp
    use fill_ctparameters, only: init_CTparameters_mdl
    use fill_couplings,only: fill_couplings_mdl,fill_CTcouplings_mdl

    ! check if modelfile is compatible to drivers
    if (model_timestamp .ne. driver_timestamp) then
      call error_mdl("Model corrupted. Driver timestamp mismatch.", &
                    where='init_vertices')
    end if

    vertices_init = .true.
    call fill_couplings_mdl
    call init_CTparameters_mdl(.true.)
    call fill_CTcouplings_mdl
    call dict_init(1327)
    call createCurrents

  end subroutine init_vertices_mdl

    subroutine get_vertex_lorentz_rank_mdl(ty,cc,mr)
      integer, intent(in) :: ty
      logical, intent(in) :: cc(0:)
      integer, intent(out) :: mr
      select case(ty)
      case(11)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(15)
        if (cc(1) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(16)
        if (cc(2) .or. cc(4)) then
          mr = 2
        else if(cc(0) .or. cc(1) .or. cc(3)) then
          mr = 1
        else
          mr = 0
        end if 
      case(18)
        if (cc(2) .or. cc(8)) then
          mr = 2
        else if(cc(1) .or. cc(3) .or. cc(9) .or. cc(10)) then
          mr = 1
        else
          mr = 0
        end if 
      case(19)
        if (cc(1) .or. cc(3) .or. cc(4) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(21)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(23)
        if (cc(1) .or. cc(2)) then
          mr = 2
        else if(cc(0) .or. cc(3)) then
          mr = 1
        else
          mr = 0
        end if 
      case(33)
        if (cc(0) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(42)
        if (cc(1) .or. cc(3) .or. cc(4) .or. cc(5) .or. cc(7) .or. cc(9) .or.  &
        cc(10)) then
          mr = 2
        else if(cc(0) .or. cc(2) .or. cc(6) .or. cc(8)) then
          mr = 1
        else
          mr = 0
        end if 
      case(51)
        if (cc(1) .or. cc(2)) then
          mr = 2
        else if(cc(0) .or. cc(3)) then
          mr = 1
        else
          mr = 0
        end if 
      case(62)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(68)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(75)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(87)
        if (cc(3) .or. cc(4)) then
          mr = 1
        else
          mr = 0
        end if 
      case(91)
        if (cc(2) .or. cc(4)) then
          mr = 2
        else if(cc(0) .or. cc(1) .or. cc(3)) then
          mr = 1
        else
          mr = 0
        end if 
      case(93)
        if (cc(2) .or. cc(4)) then
          mr = 2
        else if(cc(0) .or. cc(1) .or. cc(3)) then
          mr = 1
        else
          mr = 0
        end if 
      case(95)
        if (cc(1) .or. cc(3) .or. cc(4) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(100)
        if (cc(5) .or. cc(10)) then
          mr = 2
        else if(cc(0) .or. cc(1) .or. cc(3) .or. cc(4) .or. cc(6) .or. cc(9))  &
        then
          mr = 1
        else
          mr = 0
        end if 
      case(101)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(102)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(6) .or. cc(7)) then
          mr = 2
        else if(cc(3) .or. cc(5) .or. cc(10)) then
          mr = 1
        else
          mr = 0
        end if 
      case(107)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(114)
        if (cc(2) .or. cc(4)) then
          mr = 2
        else if(cc(0) .or. cc(1) .or. cc(3)) then
          mr = 1
        else
          mr = 0
        end if 
      case(116)
        if (cc(1) .or. cc(6)) then
          mr = 2
        else if(cc(0) .or. cc(2) .or. cc(3) .or. cc(5) .or. cc(7) .or. cc(10)) &
        then
          mr = 1
        else
          mr = 0
        end if 
      case(119)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(124)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(125)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(126)
        if (cc(0) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(131)
        if (cc(3) .or. cc(4)) then
          mr = 1
        else
          mr = 0
        end if 
      case(132)
        if (cc(1) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(135)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(147)
        if (cc(1) .or. cc(3) .or. cc(4) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(148)
        if (cc(4) .or. cc(8)) then
          mr = 2
        else if(cc(3) .or. cc(5) .or. cc(9) .or. cc(10)) then
          mr = 1
        else
          mr = 0
        end if 
      case(150)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(152)
        if (cc(1) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(153)
        if (cc(1) .or. cc(2) .or. cc(3) .or. cc(5) .or. cc(7) .or. cc(9) .or.  &
        cc(10)) then
          mr = 2
        else if(cc(0) .or. cc(4) .or. cc(6) .or. cc(8)) then
          mr = 1
        else
          mr = 0
        end if 
      case(154)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(156)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(160)
        if (cc(3) .or. cc(4)) then
          mr = 1
        else
          mr = 0
        end if 
      case(162)
        if (cc(2) .or. cc(8)) then
          mr = 2
        else if(cc(1) .or. cc(3) .or. cc(9) .or. cc(10)) then
          mr = 1
        else
          mr = 0
        end if 
      case(164)
        if (cc(0) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(169)
        if (cc(1) .or. cc(6)) then
          mr = 2
        else if(cc(0) .or. cc(2) .or. cc(3) .or. cc(4) .or. cc(7) .or. cc(10)) &
        then
          mr = 1
        else
          mr = 0
        end if 
      case(172)
        if (cc(1) .or. cc(3) .or. cc(4) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(174)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(177)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(178)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(180)
        if (cc(1) .or. cc(3) .or. cc(4) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case(181)
        if (cc(1) .or. cc(2) .or. cc(3) .or. cc(5) .or. cc(6) .or. cc(8) .or.  &
        cc(10)) then
          mr = 2
        else if(cc(0) .or. cc(4) .or. cc(7) .or. cc(9)) then
          mr = 1
        else
          mr = 0
        end if 
      case(183)
        if (cc(0) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(198)
        if (cc(3) .or. cc(4)) then
          mr = 1
        else
          mr = 0
        end if 
      case(201)
        if (cc(0) .or. cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(204)
        if (cc(1) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(221)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(227)
        if (cc(0) .or. cc(3)) then
          mr = 2
        else if(cc(1)) then
          mr = 1
        else
          mr = 0
        end if 
      case(231)
        if (cc(1) .or. cc(2)) then
          mr = 1
        else
          mr = 0
        end if 
      case(232)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(233)
        if (cc(0) .or. cc(1) .or. cc(2) .or. cc(3) .or. cc(4)) then
          mr = 2
        else
          mr = 0
        end if 
      case(235)
        if (cc(0)) then
          mr = 1
        else
          mr = 0
        end if 
      case(238)
        if (cc(1) .or. cc(5)) then
          mr = 1
        else
          mr = 0
        end if 
      case default
        mr = 0
      end select 
    end subroutine get_vertex_lorentz_rank_mdl
end module class_vertices
